""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import os
import sys
import json
import time
from datetime import date,datetime
from helpers.cmd_utils import CmdUtils
from framework.utility.dual_output_handler import DualOutputHandler
from framework.utility.task_utils import TaskUtilities
from helpers.logger import Logger
from constants import *
from framework.utility import messages
from helpers.task_file_generator import TaskFileGenerator
from helpers.shell_file_generator import ShellFileGenerator
from helpers.utils import Utilities


class Framework(object):
    def __init__(self) -> None:
        self.cmd_line_utilities = CmdUtils()
        self.task_utilities = TaskUtilities()
        self.utilities = Utilities()
        self.number_of_args = len(sys.argv)
        self.current_version = self.task_utilities.get_current_version()
        self.logger = Logger.get_logger(__name__)

    def print_data(self, phase, phase_tasks):
        unregistered_tasks = phase_tasks["unregistered_tasks"]
        sorted_dict_of_phase_tasks = phase_tasks["sorted_dict_of_phase_tasks"]
        if unregistered_tasks:
            self.task_utilities.print_justified_text(
                messages.UNREGISTERED_TASKS)
            print(unregistered_tasks)
        if sorted_dict_of_phase_tasks:
            self.task_utilities.print_justified_text(
                messages.LIST_OF_TASKS.format(phase))
            for index, task in sorted_dict_of_phase_tasks.items():
                print("{:<4}{:<50}".format(str(index)+'.',task))

    def _function_code(self, version, phase_name, operation, dry_run=False, skip_task_download=False, status_file=None):
        try:
            phase = PHASE_MAP_DICT[phase_name]
            phase_tasks = {"sorted_dict_of_phase_tasks": {},
                           "unregistered_tasks": []}
            # Following are some checks for version. 
            if phase_name in CHECK_READINESS_PHASES:
                target_version = version
                result = self.utilities.version_compare(
                    self.current_version, target_version)
                if result == 0:
                    # target version is same as current version
                    msg = messages.LOWER_OR_SAME_VERSION_PROVIDED.format(
                        self.current_version, version, "same as")
                    self.logger.error("ERROR: {}".format(msg))
                    self.task_utilities.print_justified_text(msg)
                    return False

                if result == 1:
                    # target version is lower than current version
                    msg = messages.LOWER_OR_SAME_VERSION_PROVIDED.format(
                        self.current_version, version, "less than")
                    self.logger.error("ERROR: {}".format(msg))
                    self.task_utilities.print_justified_text(msg)
                    return False

                is_valid_upgrade, msg = self.utilities.check_upgrade_supported(
                    self.current_version, target_version)
                if not is_valid_upgrade:
                    self.logger.error("ERROR: {}".format(msg))
                    self.task_utilities.print_justified_text(msg)
                    return False
            
            if not skip_task_download:
                self.task_utilities.download_tasks_from_repo(version)
            
            sorted_dict_of_phase_tasks, unregistered_tasks = self.task_utilities.identify_tasks(
                phase)
            if unregistered_tasks:
                phase_tasks["unregistered_tasks"] = unregistered_tasks
                if operation == 'execute':
                    msg = messages.UNREGISTERED_TASKS.format(
                        unregistered_tasks)
                    self.logger.error(msg)
                    self.task_utilities.print_justified_text(msg)
                    sys.exit()
            if sorted_dict_of_phase_tasks:
                if operation == 'list':
                    phase_tasks["sorted_dict_of_phase_tasks"] = sorted_dict_of_phase_tasks
                    return phase_tasks

                # Following code will create report-<version>.json & step_result-<version>.json files if not present with metadata in it.
                # If present, will update metadata in both files.
                current_datetime = datetime.utcnow()
                current_datetime = current_datetime.strftime("%d/%m/%Y, %H:%M:%S")
                metadata = {"Current FortiSOAR version" : f"{self.current_version}"}
                file_data={}
                if not os.path.isfile(REPORT_FILE.format(version)):
                    with open(REPORT_FILE.format(version), 'w') as file:
                        metadata["File creation time"] = f"{current_datetime}"
                        metadata["File modification time"] = ""
                        file_data["metadata"] = metadata
                        file.write(json.dumps(file_data))
                    os.system(
                        f"chown nginx:nginx {REPORT_FILE.format(version)}")
                    os.system(f"chmod 644 {REPORT_FILE.format(version)}")
                else:
                    with open(REPORT_FILE.format(version), 'r') as file:
                        file_data = json.load(file)
                    
                    with open(REPORT_FILE.format(version), 'w') as file:
                        file_data["metadata"]["Current FortiSOAR version"] = f"{self.current_version}"
                        file_data["metadata"]["File modification time"] = f"{current_datetime}"
                        file_data = json.dumps(file_data, indent=4, separators=(',', ': '))
                        file.write(file_data)

                if not os.path.isfile(STEP_RESULT_FILE.format(version)):
                    with open(STEP_RESULT_FILE.format(version), 'w') as file:
                        metadata["File creation time"] = current_datetime
                        metadata["File modification time"] = ""
                        file_data["metadata"] = metadata
                        file.write(json.dumps(file_data))
                    os.system(
                        f"chown nginx:nginx {STEP_RESULT_FILE.format(version)}")
                    os.system(f"chmod 644 {STEP_RESULT_FILE.format(version)}")
                else:
                    with open(STEP_RESULT_FILE.format(version), 'r') as file:
                        file_data = json.load(file)
                    
                    with open(STEP_RESULT_FILE.format(version), 'w') as file:
                        file_data["metadata"]["Current FortiSOAR version"] = f"{self.current_version}"
                        file_data["metadata"]["File modification time"] = f"{current_datetime}"
                        file_data = json.dumps(file_data, indent=4, separators=(',', ': '))
                        file.write(file_data)

                # This will execute all task files under prvided phase
                self.task_utilities.execute_tasks(sorted_dict_of_phase_tasks, phase, dry_run, status_file)
                return True
            else:
                self.task_utilities.print_justified_text(
                    "No tasks in {} phase.".format(phase_name))
        except KeyError as ke:
            self.logger.exception("KeyError occured : {}".format(ke))
            self.task_utilities.print_justified_text(
                messages.PROVIDE_PHASE_FROM_LIST)
            self.list_phases()
        except Exception as ex:
            self.logger.exception(ex)

    def check_readiness(self, version):
        try:
            self.utilities.check_for_version_argument(version)
            current_process_pid = os.getpid()
            result = self.utilities.is_command_running(
                'check-readiness', CHECK_READINESS_STATUS_FILE)
            if result[0]:
                err_msg = messages.ERROR_OCCURED.format(
                    "execution of check-readiness command", messages.COMMAND_RUNNING.format(result[1]))
                self.logger.error(err_msg)
                sys.exit(err_msg)

            self.utilities.update_command_status(version, current_process_pid, None, None, 'started', CHECK_READINESS_STATUS_FILE)

            for phase_name in CHECK_READINESS_PHASES:
                result = self._function_code(
                    version, phase_name, 'execute', True, False, CHECK_READINESS_STATUS_FILE)

            self.utilities.update_command_status(version, current_process_pid, None, None, 'ended', CHECK_READINESS_STATUS_FILE)

            if not result:
                sys.exit()
            if result and os.path.isfile(f"{FRAMEWORK_PATH}/outputs/report-{version}.json"):
                # '\033[33;1;5mNote:\033[0m' This code will display message 'Note' in highlighted format on cli.
                self.task_utilities.print_justified_text(
                    "\033[33;1;5mNote:\033[0m "+messages.REPORT_GENERATED.format(f'report-{version}.json', f"{FRAMEWORK_PATH}/outputs/"))
        except Exception as ex:
            self.logger.exception(
                messages.ERROR_OCCURED.format("check readiness", ex))
            self.utilities.update_command_status(version, current_process_pid, None, None, 'ended', CHECK_READINESS_STATUS_FILE)
            sys.exit(messages.REFER_LOG)

    def list_phases(self):
        self.task_utilities.print_justified_text("List of phases :")
        for phase in LIST_PHASES:
            print("{:<5}{:<1}{:<10}".format("", "- ", phase))

    def list_phase_tasks(self, version, phase):
        try:
            self.utilities.check_for_version_argument(version)
            result = self._function_code(version, phase, 'list')
            # 'result' will contain dictionary of phase tasks if supported FortiSOAR version is provided, else it will have boolean value false.
            if isinstance(result, bool) and not result:
                sys.exit()
            self.print_data(phase, result)
        except Exception as ex:
            self.logger.exception(messages.ERROR_OCCURED.format(
                "listing phase tasks", ex))
            sys.exit(messages.REFER_LOG)

    def execute_phase(self, version, phase, skip_task_download):
        try:
            self.utilities.check_for_version_argument(version)
            current_process_pid = os.getpid()
            result = self.utilities.is_command_running(
                'execute-phase', EXECUTE_PHASE_STATUS_FILE)
            if result[0]:
                err_msg = messages.ERROR_OCCURED.format(
                    "execution of execute-phase command", messages.COMMAND_RUNNING.format(result[1]))
                self.logger.error(err_msg)
                sys.exit(err_msg)

            current_version = int(self.current_version.replace('.',''))
            target_version = int(version.replace('.',''))
            if phase == 'postupgrade':
                if current_version < target_version:
                    self.task_utilities.print_justified_text(messages.POST_UPGRADE_EXECUTION_RESTRICTION.format(version,self.current_version))
                    sys.exit()
                if os.path.exists(f"/opt/fsr-elevate/elevate/outputs/step_result-{version}.json"):
                    with open(f"/opt/fsr-elevate/elevate/outputs/step_result-{version}.json", 'r') as file:
                        step_result_data = json.load(file)
                        if step_result_data == {}:
                            msg = "{}\n{}".format(messages.STEP_RESULT_EMPTY.format(version),messages.EXECUTE_PREUPGRADE_PHASE.format(version))
                            self.logger.error(messages.STEP_RESULT_EMPTY.format(version))
                            self.task_utilities.print_justified_text(msg)
                            sys.exit()
                else:
                    msg = "{}\n{}".format(messages.STEP_RESULT_NOT_EXISTS.format(version),messages.EXECUTE_PREUPGRADE_PHASE.format(version))
                    self.logger.error(messages.STEP_RESULT_NOT_EXISTS.format(version))
                    self.task_utilities.print_justified_text(msg)
                    sys.exit()
            
            status_task = None
            status_phase = None
            if os.path.isfile(EXECUTE_PHASE_STATUS_FILE):
                status_data = self.utilities.read_status_file(EXECUTE_PHASE_STATUS_FILE)
                if status_data:
                    status_data_task = status_data.get('task',None)
                    status_data_phase = status_data.get('phase',None)
                    status_data_status = status_data.get('status',None)
                    if phase == status_data_phase and status_data_task and status_data_status == 'started':
                        status_task = status_data['task']
                        status_phase = status_data['phase']
            
            self.utilities.update_command_status(version, current_process_pid, status_phase, status_task, 'started', EXECUTE_PHASE_STATUS_FILE)

            result = self._function_code(version, phase, 'execute', False, skip_task_download, EXECUTE_PHASE_STATUS_FILE)

            self.utilities.update_command_status(version, current_process_pid, None, None, 'ended', EXECUTE_PHASE_STATUS_FILE)

            if not result:
                sys.exit()
        except Exception as ex:
            msg_sub_string = "executing {} phase".format(phase)
            self.logger.exception(
                messages.ERROR_OCCURED.format(msg_sub_string, ex))
            self.utilities.update_command_status(version, current_process_pid, None, None, 'ended', EXECUTE_PHASE_STATUS_FILE)
            sys.exit(messages.REFER_LOG)

    def execute(self, version):
        try:
            self.utilities.check_for_version_argument(version)
            current_process_pid = os.getpid()
            result = self.utilities.is_command_running(
                'execute', EXECUTE_STATUS_FILE)
            if result[0]:
                err_msg = messages.ERROR_OCCURED.format(
                    "execution of execute command", messages.COMMAND_RUNNING.format(result[1]))
                self.logger.error(err_msg)
                sys.exit(err_msg)
            
            today_date = date.today()
            time_epoch = int(time.time())
            log_file = UPGRADE_LOG_FILE.format(version,today_date,time_epoch)
            log_file_obj = open(log_file,'a')
            original_stdout = sys.stdout
            # Redirect standard output to both CLI and a log file using DualOutputHandler
            sys.stdout = DualOutputHandler(sys.stdout, log_file_obj)
                
            for phase_name in PHASE_MAP_DICT.keys():
                # To handle very first execution of command/phase
                status_data = None
                status_task = None
                status_phase = None
                status = 'started'
                
                if os.path.isfile(EXECUTE_STATUS_FILE):
                    status_data = self.utilities.read_status_file(EXECUTE_STATUS_FILE)
                    if status_data:
                        status_data_task = status_data.get('task',None)
                        status_data_phase = status_data.get('phase',None)
                        status_data_status = status_data.get('status',None)

                        # This will ensure to skip execution of already executed phases
                        if status_data_phase and phase_name != status_data_phase and status_data_status == 'started':
                            continue

                        if status_data_task and status_data_status == 'started':
                            # To handle resume of execution of command from point of failure
                            status_task = status_data_task
                            status_phase = status_data_phase
                            status = status_data_status

                self.utilities.update_command_status(version, current_process_pid, status_phase, status_task, status, EXECUTE_STATUS_FILE)
                result = self._function_code(version, phase_name, 'execute', False, False, EXECUTE_STATUS_FILE)
            
            # Restore the original standard output
            sys.stdout = original_stdout
            log_file_obj.close()

            self.utilities.update_command_status(version, current_process_pid, None, None, 'ended', EXECUTE_STATUS_FILE)

            if not result:
                sys.exit()
        except Exception as ex:
            self.logger.exception(messages.ERROR_OCCURED.format("FSR upgrade", ex))
            self.utilities.update_command_status(version, current_process_pid, None, None, 'ended', EXECUTE_STATUS_FILE)
            sys.exit(messages.REFER_LOG)

    def extract_zip(self, zip_path):
        try:
            self.task_utilities.extract_zip(zip_path)
            zip_file = zip_path.split('/')[-1]
            self.task_utilities.print_justified_text(
                messages.EXTRACTED_SUCCESSFULLY.format(zip_file, f"{FRAMEWORK_PATH}/workflow"))
        except Exception as ex:
            self.logger.exception(messages.ERROR_OCCURED.format(
                "extracting files from zip", ex))
            sys.exit(messages.REFER_LOG)

    def create_task(self, phase, task_name, cls_name):
        task_generator = TaskFileGenerator()
        task_generator.verify_arguments(phase, task_name, cls_name)
        task_generator.create_task()

    def create_shell_script(self, phase, shell_script_name):
        shell_file_generator = ShellFileGenerator()
        shell_file_generator.verify_arguments(phase, shell_script_name)
        shell_file_generator.create_script()
