""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import os
import sys
import json
import zipfile
import pathlib
import re
import importlib
import inspect
import textwrap
from helpers.logger import Logger
from constants import *
from framework.utility import messages
from helpers.utils import Utilities
import os
import requests
from cshmac.requests import HmacAuth

KEY_PATH = '/opt/cyops/configs/keys/'

class TaskUtilities(object):
    def __init__(self) -> None:
        self.logger = Logger.get_logger(__name__)
        self.utilities = Utilities()
        self.product_yum_server = os.environ['product_yum_server'] if 'product_yum_server' in os.environ else REPO_URL

    def _show_start_or_end_msg(self, msg, position=None):
        if position == 'start':
            msg = "Started executing "+msg
        if position == 'end':
            msg = "Completed execution of "+msg
        border = "="*(MSG_WIDTH+SPACE_BEFORE_MSG)
        final_msg = (" "*SPACE_BEFORE_MSG)+msg
        print("\n{}\n{}\n{}".format(border, final_msg, border))

    def download_tasks_from_repo(self, target_fsr_version):
        try:
            zip_file = "workflow.zip"
            zip_file_url = REPO_PATH.format(
                self.product_yum_server, target_fsr_version, target_fsr_version)
            if 'https' in self.product_yum_server or 'http' in self.product_yum_server:
                if self.product_yum_server.endswith('/'):
                    zip_file_url = str(
                        self.product_yum_server)+f"{target_fsr_version}/elevate/elevate-{target_fsr_version}.zip"
                else:
                    zip_file_url = str(
                        self.product_yum_server)+f"/{target_fsr_version}/elevate/elevate-{target_fsr_version}.zip"

            response_dict = self.utilities.send_get_request(zip_file_url)
            if response_dict[0] != 200:
                self.logger.error(response_dict[1])
                sys.exit(response_dict[1])
            open(zip_file, 'wb').write(response_dict[1])

            checksum_file = "checksum.txt"
            checksum_url = CHECKSUM_PATH.format(
                self.product_yum_server, target_fsr_version, target_fsr_version)
            if 'https' in self.product_yum_server or 'http' in self.product_yum_server:
                if self.product_yum_server.endswith('/'):
                    checksum_url = str(
                        self.product_yum_server)+f"{target_fsr_version}/elevate/elevate-{target_fsr_version}-checksum.txt"
                else:
                    checksum_url = str(
                        self.product_yum_server)+f"/{target_fsr_version}/elevate/elevate-{target_fsr_version}-checksum.txt"

            response_dict = self.utilities.send_get_request(checksum_url)
            if response_dict[0] != 200:
                self.logger.error(response_dict[1])
                sys.exit(response_dict[1])
            open(checksum_file, 'wb').write(response_dict[1])

            downloaded_file_hash = self.utilities.get_file_hash(zip_file)
            with open(checksum_file, 'r') as cf:
                file_hash_provided = cf.read().split(' ')[1]
            if downloaded_file_hash == file_hash_provided:
                self.extract_zip()
                repo_update_rpm = f"cyops-repo-update-{target_fsr_version}.el8.x86_64.rpm"
                if os.path.isfile(f"{FRAMEWORK_PATH}/{repo_update_rpm}"):
                    os.system(
                        f"chown nginx:nginx {FRAMEWORK_PATH}/{repo_update_rpm}")

            else:
                self.logger.error(messages.CHECKSUM_FAILED)
                os.remove(zip_file)
                os.remove(checksum_file)
                sys.exit(messages.CHECKSUM_FAILED)

            os.remove(zip_file)
            os.remove(checksum_file)
        except Exception as ex:
            self.logger.exception(ex)
            self.print_justified_text(messages.ERROR_OCCURED.format(
                'downloading of tasks from repo', ex))

    def execute_tasks(self, sorted_dict_of_phase_tasks, phase, dry_run, status_file=None):
        current_process_pid = os.getpid()
        version = self.get_target_version()
        phase_name = list(PHASE_MAP_DICT.keys())[list(PHASE_MAP_DICT.values()).index(phase)]
        self.logger.info(messages.PHASE_START_MSG.format(phase_name))
        self._show_start_or_end_msg(phase_name+" phase", "start")
        try:
            all_tasks_from_phase = list(sorted_dict_of_phase_tasks.values())
            tasks_to_be_executed = None
            executed_tasks = None
            if status_file:
                with open(status_file,'r') as file:
                    file_data = json.load(file)
                    if file_data:
                        file_data_status = file_data.get('status',None)
                        file_data_task = file_data.get('task',None)
                        if file_data_status == 'started':
                            task = file_data_task
                            if task:
                                task_index = all_tasks_from_phase.index(task)
                                tasks_to_be_executed = all_tasks_from_phase[task_index:]
                                executed_tasks = all_tasks_from_phase[:task_index]
            tasks_to_be_executed = tasks_to_be_executed if tasks_to_be_executed else all_tasks_from_phase
            if executed_tasks:
                for executed_task in executed_tasks:
                    self.print_justified_text(messages.TASK_SUCCESSFUL_EXECUTION.format(executed_task))
                msg = messages.EXECUTION_RESUME_MSG.format(phase_name,task)
                self.print_justified_text("\n" + msg)
            for task in tasks_to_be_executed:
                module_to_import = MODULE_TO_IMPORT.format(phase, task)
                module_import = importlib.import_module(module_to_import)
                task_class = inspect.getmembers(module_import, inspect.isclass)
                for class_name, class_type in task_class:
                    if class_name == 'Tasks':
                        continue
                    if dir(class_type).count('validate') > 0 and dry_run:
                        obj = class_type()
                        if obj.is_supported():
                            obj.validate()
                        else:
                            not_supported_msg = "{} is not supported.".format(
                                task)
                            self.logger.info(not_supported_msg)

                    if dir(class_type).count('execute') > 0 and not dry_run:
                        obj = class_type()
                        if status_file:
                            self.utilities.update_command_status(version, str(current_process_pid), phase_name, task, 'started', status_file)
                        
                        if not obj.is_supported():
                            continue
                        self.print_justified_text(obj.get_description())
                        if obj.validate():
                            obj.execute()
                            if status_file:
                                self.utilities.update_command_status(version, str(current_process_pid), phase_name, task, 'ended', status_file)
                        else:
                            self.logger.error(
                                messages.VALIDATE_FAILED.format(phase_name))
                            self._show_start_or_end_msg(phase_name+" phase", "end")
                            self.logger.info(
                                messages.PHASE_END_MSG.format(phase_name))
                            self.print_justified_text(messages.VALIDATE_FAILED.format(
                                phase_name)+"\n"+messages.REFER_LOG)
                            sys.exit()
        except Exception as ex:
            err_msg = messages.ERROR_OCCURED.format(
                f"executing {phase_name} phase", ex)
            self.logger.exception(err_msg)
            self.print_justified_text(err_msg)
            self.print_justified_text(messages.REFER_LOG)
            sys.exit()

        self._show_start_or_end_msg(phase_name+" phase", "end")
        self.logger.info(messages.PHASE_END_MSG.format(phase_name))

    def identify_tasks(self, phase_name) -> tuple:
        try:
            dict_of_phase_tasks = {}
            unregistered_tasks = []
            scan_dir_path = '{}/workflow/{}'.format(
                FRAMEWORK_PATH, phase_name)
            if not os.path.exists(scan_dir_path):
                self.logger.warning(PHASE_DIR_NOT_EXISTS.format(phase_name))
                return ({}, unregistered_tasks)
            dir_content = os.scandir(scan_dir_path)
            for item in dir_content:
                if not item.name.endswith('.py'):
                    continue
                if item.is_file():
                    if len(re.findall(r"(?=("+'|'.join(UNWANTED_FILES_DIR)+r"))", str(item))) != 0:
                        continue
                    task_sequence = int(item.name.split('.')[0].split('_')[0])
                    task_module = item.name.split('.')[0]
                    module_to_import = MODULE_TO_IMPORT.format(
                        phase_name, task_module)
                    module_import = importlib.import_module(module_to_import)
                    task_class = inspect.getmembers(
                        module_import, inspect.isclass)
                    list_of_classes = [
                        class_name for class_name, class_type in task_class]
                    if 'Tasks' not in list_of_classes:
                        unregistered_tasks.append(task_module)
                    dict_of_phase_tasks[task_sequence] = task_module
            task_sequence_list = list(dict_of_phase_tasks.keys())
            task_sequence_list.sort()
            sorted_dict_of_phase_tasks = {
                seq_no: dict_of_phase_tasks[seq_no] for seq_no in task_sequence_list}
            sorted_dict_of_phase_tasks = {
                key: val for key, val in sorted_dict_of_phase_tasks.items() if val not in unregistered_tasks}
            new_key = 1
            new_sorted_dict_of_phase_tasks = {}
            for old_key in list(sorted_dict_of_phase_tasks):
                new_sorted_dict_of_phase_tasks[new_key] = sorted_dict_of_phase_tasks.pop(
                    old_key)
                new_key += 1
            return new_sorted_dict_of_phase_tasks, unregistered_tasks
        except ValueError as ve:
            self.logger.exception(ve)
            return ({}, unregistered_tasks)
        except Exception as exp:
            self.logger.exception(exp)
            raise Exception(str(exp))

    def extract_zip(self, zip_path=None) -> None:
        try:
            zip_file = zip_path if zip_path else str(
                pathlib.Path.cwd()) + '/workflow.zip'

            # Get a list of all items (files and directories) in the "/opt/fsr-elevate/elevate" directory before zip extraction
            all_items = os.listdir(FRAMEWORK_PATH)
            # Filter only directories from the list
            dirs_before_zip_extraction = [item for item in all_items if os.path.isdir(
                os.path.join(FRAMEWORK_PATH, item))]

            with zipfile.ZipFile(zip_file, 'r') as zip_ref:
                zip_ref.extractall(FRAMEWORK_PATH)

            # Get a list of all items (files and directories) in the "/opt/fsr-elevate/elevate" directory after zip extraction
            all_items = os.listdir(FRAMEWORK_PATH)
            # Filter only directories from the list
            dirs_after_zip_extraction = [item for item in all_items if os.path.isdir(
                os.path.join(FRAMEWORK_PATH, item))]

            list_diff = set(dirs_after_zip_extraction).difference(
                dirs_before_zip_extraction)
            if list_diff:
                if zip_path:
                    zip_file = zip_path.split("/")[-1]
                    msg = messages.PROVIDED_ZIP_EXTRACTION_ERROR.format(
                        zip_file)
                else:
                    target_version = self.get_target_version()
                    msg = messages.REPO_ZIP_EXTRACTION_ERROR.format(
                        target_version) + "Please contact administrator."
                self.logger.error(msg)
                self.print_justified_text(msg)
                sys.exit()
        except FileNotFoundError:
            sys.exit(messages.NO_FILE_OR_DIR.format(zip_path))
        except Exception as exp:
            self.logger.exception(exp)
            self.print_justified_text(messages.ERROR_OCCURED.format(
                'extracting files from zip', exp))

    def check_task_number_exists(self, phase_name, task_file_name):
        try:
            new_task_seq = int(task_file_name.split('_')[0])
            list_of_task_seq = []
            sorted_dict_of_phase_tasks, unregistered_tasks = self.identify_tasks(
                phase_name)
            for task_seq, task_name in sorted_dict_of_phase_tasks.items():
                list_of_task_seq.append(int(task_name.split('_')[0]))
            for task_name in unregistered_tasks:
                list_of_task_seq.append(int(task_name.split('_')[0]))

            list_of_task_seq.sort()
            if new_task_seq in list_of_task_seq:
                return True, new_task_seq, list_of_task_seq
            else:
                return False, new_task_seq, list_of_task_seq
        except ValueError as ve:
            return str(ve)

    def is_pascal_case(self, string):
        return bool(re.match(PASCAL_CASE_PATTERN, string)) and not string.isupper()

    def print_justified_text(self, input_text):
        if isinstance(input_text, str):
            new_line_chr_pattern = r'\n'
            matches = re.findall(new_line_chr_pattern, input_text)
            if len(matches) > 0:
                text_list = input_text.split('\n')
                for text in text_list:
                    justified_text = textwrap.fill(text, width=MSG_WIDTH)
                    print(justified_text)
            else:
                justified_text = textwrap.fill(input_text, width=MSG_WIDTH)
                print(justified_text)
        else:
            print(input_text)

    def store_result_to_json(self, json_file_path, phase_name, task_name, key=None, data=None) -> None:
        if os.path.exists(json_file_path):
            with open(json_file_path, 'r') as step_result_json_file:
                json_object = json.load(step_result_json_file)
            if phase_name is None and task_name is None and key and data:
                # Just key and value in json file
                json_object[key] = data
            elif phase_name in json_object:
                # This will update value of particular key
                if task_name in json_object[phase_name] and key:
                    json_object[phase_name][task_name][key] = data
                else:
                    if key:
                        data = {key: data}
                    json_object[phase_name][task_name] = data
            else:
                if key:
                    data = {key: data}
                json_object[phase_name] = {task_name: data}
            with open(json_file_path, 'w') as step_result_json_file:
                json_object = json.dumps(
                    json_object, indent=4, separators=(',', ': '))
                step_result_json_file.write(json_object)
        else:
            if key:
                data = {key: data}
            step_result = {phase_name: {task_name: data}}
            json_object = json.dumps(
                step_result, indent=4, separators=(',', ': '))
            file = open(json_file_path, 'x')
            file.write(json_object)

    def get_current_version(self) -> str:
        try:
            with open(VERSION_FILE, 'r') as version_file:
                json_object = json.load(version_file)
                current_version = json_object['version'].split('-')[0]
            return current_version
        except FileNotFoundError:
            raise FileNotFoundError(
                f"{messages.NO_FILE_OR_DIR.format(VERSION_FILE)}")
        except json.JSONDecodeError as jsnerr:
            raise json.JSONDecodeError(
                f"{messages.JSON_PARSE_ERROR.format(jsnerr)}")

    def get_target_version(self) -> str:
        if len(sys.argv) > 1 and '--version' in sys.argv:
            version_arg_index = sys.argv.index('--version')+1
            target_upgrade_version = sys.argv[version_arg_index]
            return target_upgrade_version
        return ""
    
    def _make_cyops_request(self, endpoint:str, method:str, payload=None):
        try:
            private_key, public_key = self._get_appliance_keys()    
            if method.lower() in ['delete', 'get']:
                auth = HmacAuth(endpoint, method, public_key, private_key, public_key)
            else:
                auth = HmacAuth(endpoint, method, public_key, private_key, json.dumps(payload))
            response = requests.request(method, endpoint, auth=auth, verify=False)
            return response
        except Exception as ex:
            self.logger.exception(ex)
            self.print_justified_text(ex)
            return {}
    
    def _get_appliance_keys(self) -> tuple:
        with open(os.path.join(KEY_PATH, 'APPLIANCE_PRIVATE_KEY'), 'r') as private:
            appliance_private_key = private.read().strip()
        with open(os.path.join(KEY_PATH, 'APPLIANCE_PUBLIC_KEY'), 'r') as public:
            appliance_public_key = public.read().strip()
        return appliance_private_key, appliance_public_key
