""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

from helpers.utils import Utilities
from framework.utility import messages
from framework.utility.task_utils import TaskUtilities
from constants import *
import json
import os
import sys
import re
sys.path.append(os.path.join(os.path.dirname(__file__), "../"))


phase_name = None
script_file_name = None


class Method:
    METHOD_SIGNATURE = "{0}({1})"

    def __init__(self, method_name: str, method_args: list[str] = None, set_vargs=True,
                 set_kwargs=True, disclaimer: str = '', default_code: list[str] = None):
        self.method_name = method_name
        self.method_args = method_args
        self.set_vargs = set_vargs
        self.set_kwargs = set_kwargs
        self.disclaimer = disclaimer
        self.default_code = default_code

    def build(self, indent_level=0) -> str:
        method_definition = '\n\t'.join(self.default_code)
        method_disclaimer = '\n\t# '.join(self.disclaimer)
        if self.set_vargs:
            self.method_args.append("*args")
        if self.set_kwargs:
            self.method_args.append("**kwargs")
        args_list = ', '.join(self.method_args) if self.method_args else ''
        method_signature = self.METHOD_SIGNATURE.format(
            self.method_name, args_list) + "{"
        self.method_code = method_signature + (
            "\n\t" + "# " + method_disclaimer if method_disclaimer else '') + "\n" + "\t" + method_definition
        indented_code = ""
        for line in self.method_code.split('\n'):
            for i in range(0, indent_level):
                indented_code += '\t'
            indented_code += line + '\n'
        return indented_code + "}" + '\n'


class ShellFile:
    TAB_SPACE = "    "

    def __init__(self, sh_shebang_line, disclaimer, sh_script_name, methods=[]):
        self.shell_code = ""
        self.sh_shebang_line = sh_shebang_line
        self.disclaimer = disclaimer
        self.sh_script_name = sh_script_name
        self.methods = methods

    def build(self) -> str:
        methods = '\n'.join([method.build() for method in self.methods])
        self.shell_code = (self.sh_shebang_line + "\n" +
                           self.disclaimer) + '\n\n'+methods
        return self.shell_code.replace('\t', ShellFile.TAB_SPACE)


class ShellFileGenerator:
    def __init__(self) -> None:
        self.utils = Utilities()
        self.task_utilities = TaskUtilities()

    def verify_arguments(self, phase, script_name):
        if phase == 'upgrade':
            sys.exit(messages.GENERATOR_UPGRADE_PHASE)
        try:
            global phase_name
            phase_name = PHASE_MAP_DICT[phase]
        except KeyError:
            self.task_utilities.print_justified_text(messages.PHASE_ARG)
            for phase in PHASE_ARGUMENT:
                print("{:<5}{:<1}{:<10}".format("", "- ", phase))
            sys.exit()

        if script_name:
            if not bool(re.match(SNAKE_CASE_PATTERN, script_name)):
                self.task_utilities.print_justified_text(messages.SCRIPT_NAME_CONVENTION)
                sys.exit()
            global script_file_name
            script_file_name = script_name
        else:
            sys.exit(messages.PROVIDE_SCRIPT_NAME)

    def _load_script_template(self):
        with open(SCRIPT_TEMPLATE_PATH.format(FRAMEWORK_PATH)) as json_file:
            json_object = json.load(json_file)
            return json_object

    def _sh_from_json(self, input_json) -> str:
        """
        Converts a json sh template to script code
        The expected keys are as follows:
            1. "shebang_line": 
                * type: str
                * value: #!/usr/bin/env bash
            2. disclaimer:
                * type: str
                * value: Text to display at the top of the file. For example Copyright
            3. script_name:
                * type: str
                * value: name of script
            4. methods
                1. method_name
                    * type: str
                    * value: method name
                2. args
                    * type: list[str]
                    * value: list of arguments other than *args and **kwargs
                3. disclaimer
                    * type: str
                    * value: Text to display at the top of the method.
                4. set_vargs
                    * type boolean
                    * value: adds *args to method signature
                5. set_kwargs
                    * type boolean
                    * value: adds **kwargs to method signature
                7. default_code
                    * type: list[str]
                    * value: list of code lines to be added. Note that every element is added as is without giving any additional indent
        Args:
            input_json: input json with keys as defined as above

        Returns: Returns a generated python code as per the input json

        """
        sh_shebang_line = input_json.get("shebang_line", "#!/bin/bash")
        sh_disclaimer = input_json.get("disclaimer", None)
        sh_script_name = input_json.get("script_name", None)
        sh_methods = []
        for sh_method in input_json.get("methods", []):
            meth_method_name = sh_method.get("method_name", "dummy_method")
            meth_args = sh_method.get("args", None)
            meth_disclaimer = sh_method.get("disclaimer", None)
            meth_set_vargs = sh_method.get('set_vargs', None)
            meth_set_kwargs = sh_method.get('set_kwargs', None)
            meth_default_code = sh_method.get('default_code', None)
            sh_method_obj = Method(method_name=meth_method_name,
                                   method_args=meth_args,
                                   set_vargs=meth_set_vargs,
                                   set_kwargs=meth_set_kwargs,
                                   disclaimer=meth_disclaimer,
                                   default_code=meth_default_code)
            sh_methods.append(sh_method_obj)
        sh_object = ShellFile(sh_shebang_line=sh_shebang_line, disclaimer=sh_disclaimer,
                              sh_script_name=sh_script_name, methods=sh_methods)
        return sh_object.build()

    def _create_and_write_shell_file(self, file_path, file_code, file_name, dir_name):
        try:
            file = open(file_path, 'x')
            file.write(file_code)
            file.close()
            return True
        except FileExistsError as FE:
            self.task_utilities.print_justified_text(messages.FILE_EXISTS.format(
                file_name + '.sh', f"{dir_name}/scripts"))
            return False

    def create_script(self):
        json_object = self._load_script_template()
        script_file_code = self._sh_from_json(json_object)
        script_name = script_file_name.lower()
        if script_name.endswith('.sh'):
            script_name = script_name.split('.')[0]
        phase_dir = "{}/workflow/{}".format(
            FRAMEWORK_PATH, phase_name)
        if not self.utils.check_dir_exists(phase_dir):
            os.mkdir(phase_dir)
        script_dir = "{}/workflow/{}/scripts".format(
            FRAMEWORK_PATH, phase_name)
        script_file = "{}/workflow/{}/scripts/{}.sh".format(
            FRAMEWORK_PATH, phase_name, script_name)
        if not self.utils.check_dir_exists(script_dir):
            os.mkdir(script_dir)
        task_got_created = self._create_and_write_shell_file(
            script_file, script_file_code, script_name, phase_name)
        if task_got_created:
            self.task_utilities.print_justified_text(messages.SCRIPT_CREATED.format(
                script_name + '.sh', f"{FRAMEWORK_PATH}/workflow/{phase_name}/scripts"))
        else:
            self.task_utilities.print_justified_text(messages.SCRIPT_CREATION_FAILED.format(
                script_name + '.sh', f"{FRAMEWORK_PATH}/workflow/{phase_name}/scripts"))
