""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import os
import sys
import json
from constants import *
from framework.utility import messages
from framework.utility.task_utils import TaskUtilities
from helpers.utils import Utilities
import re

sys.path.append(os.path.join(os.path.dirname(__file__), "../"))

phase_name = None
task_file_name = None
class_name = None


class Method:
    METHOD_SIGNATURE = "def {0}({1}) -> {2}"

    def __init__(self, method_name: str, method_args: list[str] = None, return_type: str = None, decorators: list[str] = None, set_vargs=True,
                 set_kwargs=True, disclaimer: str = '', default_code: list[str] = None):
        self.method_name = method_name
        self.method_args = method_args
        self.return_type = return_type
        self.decorators = decorators
        self.set_vargs = set_vargs
        self.set_kwargs = set_kwargs
        self.method_code = ""
        self.method_definition = "# TODO\n\tpass"
        self.disclaimer = disclaimer
        self.default_code = default_code

    def build(self, indent_level=0) -> str:
        decorators = '@' + \
            '\n@'.join(self.decorators) if self.decorators else ''
        if phase_name == 'post_upgrade' and self.method_name == 'is_supported':
            for index in range(len(self.default_code)):
                line = self.default_code[index]
                if line.startswith('return') and line.__contains__('target_upgrade_version > current_version'):
                    line = line.replace('>','>=')
                    self.default_code[index] = line
        method_definition = '\n\t'.join(
            self.default_code) if self.default_code else self.method_definition
        if self.method_name == 'tags':
            method_definition = "return '{}'".format(phase_name)
        if self.set_vargs:
            self.method_args.append("*args")
        if self.set_kwargs:
            self.method_args.append("**kwargs")
        args_list = ', '.join(self.method_args) if self.method_args else ''
        self.return_type = self.return_type if self.return_type else None
        method_signature = self.METHOD_SIGNATURE.format(
            self.method_name, args_list, self.return_type) + ":"
        self.method_code = decorators + '\n' + method_signature + (
            "\n\t" + '"""' + self.disclaimer + '"""' if self.disclaimer else '') + "\n" + "\t" + method_definition
        indented_code = ""
        for line in self.method_code.split('\n'):
            for i in range(0, indent_level):
                indented_code += '\t'
            indented_code += line + '\n'
        return indented_code


class Class:
    BASE_CLASS_SIGNATURE = "class {0}:"
    DERIVED_CLASS_SIGNATURE = "class {0}({1}):"

    def __init__(self, class_name: str, methods: list[Method], base_classes: list[str], attributes: dict = {},
                 disclaimer: str = None):
        self.class_name = class_name
        self.generated_class = ""
        self.methods = methods
        self.base_classes = base_classes
        self.attributes = attributes
        self.disclaimer = disclaimer

    def build(self) -> str:
        methods = '\n'.join([method.build(indent_level=1)
                            for method in self.methods])
        base_classes_list = ', '.join(
            self.base_classes) if self.base_classes else ''
        class_attributes = '\n'.join(
            ['\t{0} = "{1}"'.format(key, self.attributes.get(key)) for key in self.attributes.keys()])
        if base_classes_list:
            class_signature = self.DERIVED_CLASS_SIGNATURE.format(
                self.class_name, base_classes_list)
        else:
            class_signature = self.BASE_CLASS_SIGNATURE.format(self.class_name)
        self.generated_class = class_signature + (
            "\n\t" + '"""' + self.disclaimer + '"""' if self.disclaimer else '') + "\n" + class_attributes + methods
        return self.generated_class


class PythonFile:

    TAB_SPACE = "    "

    def __init__(self, classes: list[Class], imports: list[str], global_vars: dict = {}, disclaimer: str = ''):
        self.python_code = ""
        self.classes = classes
        self.imports = imports
        self.disclaimer = disclaimer
        self.global_vars = global_vars

    def build(self) -> str:
        import_list = '\n'.join(self.imports) if self.imports else ''
        global_vars = '\n'.join(
            ['{0} = "{1}"'.format(key, self.global_vars.get(key)) for key in self.global_vars.keys()]) + '\n'
        classes = '\n\n\n'.join([cls.build() for cls in self.classes])
        self.python_code = (
            '"""' + self.disclaimer + '"""\n\n' if self.disclaimer else '') + import_list + (
            "\n\n" + global_vars + "\n\n" if len(global_vars) > 1 else '\n\n\n') + classes
        return self.python_code.replace('\t', PythonFile.TAB_SPACE)


class TaskFileGenerator:
    def __init__(self) -> None:
        self.task_utilities = TaskUtilities()
        self.utils = Utilities()

    def verify_arguments(self, phase, task, cls):
        if phase == 'upgrade':
            self.task_utilities.print_justified_text(
                messages.GENERATOR_UPGRADE_PHASE)
            sys.exit()
        try:
            global phase_name
            phase_name = PHASE_MAP_DICT[phase]
        except KeyError:
            self.task_utilities.print_justified_text(messages.PHASE_ARG)
            for phase in PHASE_ARGUMENT:
                print("{:<5}{:<1}{:<10}".format("", "- ", phase))
            sys.exit()

        if task:
            if not bool(re.match(TASK_FILE_NAME_PATTERN, task)):
                self.task_utilities.print_justified_text(
                    messages.TASK_NAME_CONVENTION)
                sys.exit()
            task_seq_no = task.split("_")[0]
            if not task_seq_no.isdigit():
                self.task_utilities.print_justified_text(
                    messages.TASK_NAME_START)
                sys.exit()
            global task_file_name
            task_file_name = task
            provided_num_present, new_task_seq, list_of_task_seq = self.task_utilities.check_task_number_exists(
                phase_name, task_file_name)
            if provided_num_present:
                self.task_utilities.print_justified_text(messages.TASK_SEQ_NUM.format(
                    new_task_seq, phase_name, list_of_task_seq))
                sys.exit()
        else:
            self.task_utilities.print_justified_text(
                messages.PROVIDE_TASK_NAME)
            sys.exit()

        if cls:
            if not self.task_utilities.is_pascal_case(cls):
                self.task_utilities.print_justified_text(messages.PASCAL_CASE)
                sys.exit()
            global class_name
            class_name = cls
        else:
            self.task_utilities.print_justified_text(
                messages.PROVIDE_CLASS_NAME)
            sys.exit()

    def _load_task_template(self):
        with open(TASK_TEMPLATE_PATH.format(FRAMEWORK_PATH)) as json_file:
            json_object = json.load(json_file)
            return json_object

    def _py_from_json(self, input_json) -> str:
        """
        Converts a json py template to python code
        The expected keys are as follows:
            1. disclaimer:
                * type: str
                * value: Text to display at the top of the file. For example Copyright
            2. imports
                * type: list[str]
                * value: List of imports
            3. globals
                * type: dict
                * value: A dict of global variables
            4. classes
                1. class_name
                    * type: str
                    * value: name of the class
                2. disclaimer
                    * type: str
                    * value: Text to display at the top of the class.
                3. base_classes
                    * type: list[str]
                    * value: list of bases classes that this class must inherit from
                4. attributes
                    * type: dict
                    * value: dictionary of class level attributes
                5. methods
                    1. method_name
                        * type: str
                        * value: method name
                    2. args
                        * type: list[str]
                        * value: list of arguments other than *args and **kwargs
                    3. disclaimer
                        * type: str
                        * value: Text to display at the top of the method.
                    4. decorators
                        * type: list[str]
                        * value: list of decorators for the given method
                    5. set_vargs
                        * type boolean
                        * value: adds *args to method signature
                    6. set_kwargs
                        * type boolean
                        * value: adds **kwargs to method signature
                    7. default_code
                        * type: list[str]
                        * value: list of code lines to be added. Note that every element is added as is without giving any additional indent
        Args:
            input_json: input json with keys as defined as above

        Returns: Returns a generated python code as per the input json

        """
        py_disclaimer = input_json.get("disclaimer", None)
        py_globals = input_json.get("globals", {})
        py_imports = input_json.get("imports", None)
        py_classes = []
        for py_class in input_json.get("classes", None):
            cls_class_name = class_name
            cls_disclaimer = py_class.get("disclaimer", None)
            cls_base_classes = py_class.get("base_classes", [])
            cls_attributes = py_class.get("attributes", {})
            cls_methods = []
            for py_method in py_class.get("methods", []):
                meth_method_name = py_method.get("method_name", "dummy_method")
                meth_args = py_method.get("args", None)
                return_type = py_method.get("return_type", None)
                meth_disclaimer = py_method.get("disclaimer", None)
                meth_decorators = py_method.get('decorators', None)
                meth_set_vargs = py_method.get('set_vargs', None)
                meth_set_kwargs = py_method.get('set_kwargs', None)
                meth_default_code = py_method.get('default_code', None)
                py_method_obj = Method(method_name=meth_method_name,
                                       method_args=meth_args,
                                       return_type=return_type,
                                       decorators=meth_decorators,
                                       set_vargs=meth_set_vargs,
                                       set_kwargs=meth_set_kwargs,
                                       disclaimer=meth_disclaimer,
                                       default_code=meth_default_code)
                cls_methods.append(py_method_obj)
            cls_object = Class(class_name=cls_class_name,
                               base_classes=cls_base_classes,
                               methods=cls_methods,
                               attributes=cls_attributes,
                               disclaimer=cls_disclaimer)
            py_classes.append(cls_object)
        py_object = PythonFile(classes=py_classes, imports=py_imports,
                               global_vars=py_globals, disclaimer=py_disclaimer)
        return py_object.build()

    def _create_and_write_task_file(self, file_path, file_code, file_name, dir_name):
        try:
            file = open(file_path, 'x')
            file.write(file_code)
            file.close()
            return True
        except FileExistsError as FE:
            self.task_utilities.print_justified_text(
                messages.FILE_EXISTS.format(file_name + '.py', dir_name))
            return False

    def create_task(self):
        json_object = self._load_task_template()
        task_file_code = self._py_from_json(json_object)
        task_name = task_file_name.lower()
        if task_name.endswith('.py'):
            task_name = task_name.split('.')[0]
        phase_dir = "{}/workflow/{}".format(FRAMEWORK_PATH, phase_name)
        task_file = "{}/workflow/{}/{}.py".format(
            FRAMEWORK_PATH, phase_name, task_name)
        if not self.utils.check_dir_exists(phase_dir):
            os.mkdir(phase_dir)
        task_got_created = self._create_and_write_task_file(
            task_file, task_file_code, task_name, phase_name)
        if task_got_created:
            self.task_utilities.print_justified_text(messages.TASK_CREATED.format(
                task_name + '.py', f"{FRAMEWORK_PATH}/workflow/{phase_name}"))
        else:
            self.task_utilities.print_justified_text(messages.TASK_CREATION_FAILED.format(
                task_name + '.py', f"{FRAMEWORK_PATH}/workflow/{phase_name}"))
