""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import os
import re
import json
import hashlib
import requests
import sys
import shlex
import subprocess
import textwrap
from helpers.logger import Logger
from constants import *
from framework.utility import messages
from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


class Utilities(object):
    def __init__(self) -> None:
        self.logger = Logger.get_logger(__name__)
        self.product_yum_server = os.environ['product_yum_server'] if 'product_yum_server' in os.environ else REPO_URL

    def check_for_version_argument(self, version_arg):
        if not version_arg:
            print(messages.VERSION_ARG_NOT_PROVIDED)
            sys.exit()

        version_pattern = re.compile(VERSION_PATTERN)
        version_non_numeric_non_dot_pattern = re.compile(
            VERSION_NON_NUMERIC_NON_DOT_CHECK_PATTERN)
        if version_non_numeric_non_dot_pattern.findall(version_arg):
            print(messages.VERSION_NON_NUMERIC_NON_DOT_FOUND)
            sys.exit()
        if not version_pattern.match(version_arg):
            print(messages.VERSION_PATTERN_NOT_FOLLOWED)
            sys.exit()

    def get_valid_upgrade_version(self, current_version) -> tuple:
        try:
            valid_upgrade_versions = []
            json_file_url = "https://" + f"{self.product_yum_server}/{current_version}/upgrade_support_versions.json"
            response = requests.get(json_file_url,)
            file_name = "upgrade_support_versions.json"
            if response.status_code and response.status_code in HTTPS_STATUS_CODES:
                if response.status_code == 200:
                    json_object = json.loads(response.content)
                    for version in json_object:
                        if version == current_version:
                            valid_upgrade_versions = json_object[version]
                            break
                    return response.status_code, valid_upgrade_versions
                msg = HTTPS_STATUS_CODES[response.status_code].format(
                    file_name)
                return response.status_code, msg
        except Exception as ex:
            self.logger.exception("ERROR: {}".format(ex))
            return 408, messages.UNABLE_TO_CONNECT.format(self.product_yum_server,self.product_yum_server)

    def check_upgrade_supported(self, current_version, target_version) -> tuple[bool, str]:
        try:
            status_code, response = self.get_valid_upgrade_version(
                current_version)
            if status_code == 200 and isinstance(response, list):
                if not response:
                    msg = messages.VERSION_DATA_UNAVAILABLE.format(
                        current_version, target_version)
                    return False, msg
                if target_version not in response:
                    msg = messages.UNSUPPORTED_VERSION_PROVIDED.format(
                        current_version, target_version, current_version, response)
                    return False, msg
                return True, ""

            if isinstance(response, str):
                return False, response
        except Exception as ex:
            self.logger.exception(f"ERROR: {ex}")
            return False, response

    def check_dir_exists(self, dir_path):
        if os.path.isdir(dir_path):
            return True
        return False

    def get_file_hash(self, file):
        sha256_hash = hashlib.sha256()
        with open(file, "rb") as f:
            # Read and update hash string value in blocks of 4K
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)
            return sha256_hash.hexdigest()

    def send_get_request(self, url, verify=False) -> list:
        response = requests.get(url, verify=verify)
        file_name = url.split('/')[-1]
        if response.status_code and response.status_code in HTTPS_STATUS_CODES:
            if response.status_code == 200:
                return [response.status_code, response.content]
            return [response.status_code, HTTPS_STATUS_CODES[response.status_code].format(file_name)]

    def is_command_running(self, command: str, status_file: str) -> list:
        if os.path.isfile(status_file):
            try:
                with open(status_file, 'r') as f:
                    data = json.load(f)
            except json.JSONDecodeError:
                commands = ["ps -ef", "grep main.py", f"grep {command}"]
                result = self.execute_commands(commands, True)
                if result:
                    return [True, command]
                return [False, command]

            if data:
                pid = data["pid"]
                commands = ["ps -ef", f"grep {pid}", f"grep {command}"]
                result = self.execute_commands(commands, True)
                if result:
                    return [True, command]
                return [False, command]
        commands = ["ps -ef", "grep main.py"]
        result = self.execute_commands(commands, True)
        processes = result.split("\n")
        process_count = 0
        commands_in_execution = []
        for process in processes:
            ps_items = list((element.strip())
                            for element in (process.split(' ')) if element)
            if process and "main.py" in process and ps_items[9] in NO_RETRIGGER_COMMANDS:
                commands_in_execution.append(ps_items[9])
                process_count += 1

        if process_count == 1:
            return [False, commands_in_execution[0]]
        return [True, commands_in_execution[0]]

    def execute_commands(self, commands: list, consume_output: bool = False):
        try:
            command_result = None
            for command in commands:
                cmd = shlex.split(command)
                if consume_output and commands.index(command) != 0:
                    result = subprocess.run(args=cmd, input=command_result.stdout, stdout=subprocess.PIPE,
                                            stderr=subprocess.PIPE, shell=False, text=True, check=True)
                else:
                    result = subprocess.run(
                        args=cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False, text=True, check=True)
                command_result = result

            if command_result and command_result.stdout:
                return command_result.stdout
            return None
        except subprocess.CalledProcessError:
            return None
        except Exception as ex:
            self.logger.exception(messages.ERROR_OCCURED.format(
                "execution of cli commands", ex))
            return None

    def update_command_status(self, version:str, pid:str, phase:str, task:str, status:str, status_file: str):
        try:
            status_data = {}
            status_data['version'] = version
            status_data['pid'] = pid
            status_data['phase'] = phase
            status_data['task'] = task
            status_data['status'] = status
            with open(status_file, 'w') as f:
                json_object = json.dumps(status_data)
                f.write(json_object)
            os.system(f"chmod 644 {status_file}")
        except FileNotFoundError:
            raise FileNotFoundError(
                f"{messages.NO_FILE_OR_DIR.format(status_file)}")

    def read_status_file(self,status_file:str):
        with open(status_file,'r') as file:
            file_data = json.load(file)
        return file_data

    # returns 0 if same,
    # returns 1 if version 1 is greater than version 2
    # returns 2 is version 1 is less than version 2
    def version_compare(self, v1, v2):
        arr1 = v1.split(".")
        arr2 = v2.split(".")
        n = len(arr1)
        m = len(arr2)
        # converts to integer from string
        arr1 = [int(i) for i in arr1]
        arr2 = [int(i) for i in arr2]

        # compares which list is bigger and fills
        # smaller list with zero (for unequal delimiters)
        if n > m:
            for i in range(m, n):
                arr2.append(0)
        elif m > n:
            for i in range(n, m):
                arr1.append(0)

        for i in range(len(arr1)):
            if arr1[i] > arr2[i]:
                return 1
            elif arr2[i] > arr1[i]:
                return 2
        return 0
