""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """

import sys
import argparse
from helpers.logger import Logger
from constants import *
from framework.framework import Framework


class Main(object):
    global_parser_var = None

    def __init__(self) -> None:
        self.logger = Logger.get_logger(__name__)
        self.framework = Framework()

    def add_check_readiness_parser(self, subparser):
        check_readiness_parser = subparser.add_parser(
            "check-readiness", help="To check whether system is ready for upgrade or not", allow_abbrev=False)
        check_readiness_parser.set_defaults(operation="check-readiness")
        check_readiness_parser.add_argument(
            "--version", action="store", type=str, required=True, help="Provide the FortiSOAR version to which the upgrade is supported")

    def add_list_phases_parser(self, subparser):
        list_phases_parser = subparser.add_parser(
            "list-phases", help="To list all the phases")
        list_phases_parser.set_defaults(operation="list-phases")

    def add_list_phase_tasks_parser(self, subparser):
        list_phase_tasks_parser = subparser.add_parser(
            "list-phase-tasks", help="To list all the tasks from a phase", allow_abbrev=False)
        list_phase_tasks_parser.set_defaults(operation="list-phase-tasks")
        list_phase_tasks_parser.add_argument(
            "--version", action="store", type=str, required=True, help="Provide the FortiSOAR version to which the upgrade is supported")
        list_phase_tasks_parser.add_argument(
            "--phase", action="store", type=str, choices=PHASE_ARGUMENT, required=True, help="Provide the name of the phase to find the relevant task files from phase")

    def add_execute_phase_parser(self, subparser):
        execute_phase_parser = subparser.add_parser(
            "execute-phase", help="To execute a particular phase", allow_abbrev=False)
        execute_phase_parser.set_defaults(operation="execute-phase")
        execute_phase_parser.add_argument(
            "--version", action="store", type=str, required=True, help="Provide the FortiSOAR version to which the upgrade is supported")
        execute_phase_parser.add_argument(
            "--phase", action="store", type=str, choices=PHASE_ARGUMENT, required=True, help="Provide the name of the phase to execute")
        execute_phase_parser.add_argument(
            "--skip-task-download", action="store_true", required=False, help=argparse.SUPPRESS)

    def add_execute_parser(self, subparser):
        execute_parser = subparser.add_parser(
            "execute", help="To upgrade FSR to provided version", allow_abbrev=False)
        execute_parser.set_defaults(operation="execute")
        execute_parser.add_argument(
            "--version", action="store", type=str, required=True, help="Provide the FortiSOAR version to which the upgrade is supported")

    def add_extract_zip_parser(self, subparser):
        extract_zip_parser = subparser.add_parser(
            "extract-zip", help="To provide tasks of a phase using zip file", allow_abbrev=False)
        extract_zip_parser.set_defaults(operation="extract-zip")
        extract_zip_parser.add_argument(
            "--zip-path", dest='zip_path', action="store", type=str, required=True, help="Provide the path where zip file is located")

    def add_create_task_parser(self, subparser):
        create_task_parser = subparser.add_parser(
            "create-task", help="To create a task file in particular phase", allow_abbrev=False)
        create_task_parser.set_defaults(operation="create-task")
        create_task_parser.add_argument(
            "--phase", action="store", type=str, choices=PHASE_ARGUMENT, required=True, help="Provide the name of the phase in which you intend to create the task file")
        create_task_parser.add_argument(
            "--task-name", dest='task_name', action="store", type=str, required=True, help="Provide task file name")
        create_task_parser.add_argument(
            "--cls-name", dest='cls_name', action="store", type=str, required=True, help="Provide class name for class which will be created in task file")

    def add_create_shell_script_parser(self, subparser):
        create_shell_script_parser = subparser.add_parser(
            "create-shell-script", help="To create a shell script file for particular phase under '<phase>/scripts'", allow_abbrev=False)
        create_shell_script_parser.set_defaults(
            operation="create-shell-script")
        create_shell_script_parser.add_argument(
            "--phase", action="store", type=str, choices=PHASE_ARGUMENT, required=True, help="Provide the name of the phase in which you intend to create the shell file")
        create_shell_script_parser.add_argument(
            "--shell-script-name", dest='shell_script_name', action="store", type=str, required=True, help="Provide shell file name")

    def define_parser(self, parser):
        subparser = parser.add_subparsers()
        self.add_check_readiness_parser(subparser)
        self.add_list_phases_parser(subparser)
        self.add_list_phase_tasks_parser(subparser)
        self.add_execute_phase_parser(subparser)
        self.add_execute_parser(subparser)
        self.add_extract_zip_parser(subparser)
        self.add_create_task_parser(subparser)
        self.add_create_shell_script_parser(subparser)

        for name, subp in subparser.choices.items():
            subp._positionals.title = 'subcommand options are'
            subp._optionals.title = 'subcommand options are'

    def main(self):
        parser = argparse.ArgumentParser(prog="csadm upgrade")
        parser._positionals.title = "subcommand options are"
        parser._optionals.title = "subcommand options are"
        self.define_parser(parser)

        global global_parser_var
        global_parser_var = parser

        args = parser.parse_args()
        if not 'operation' in dir(args):
            parser.print_help(sys.stderr)
            sys.exit(1)

        if str(args.operation) == "check-readiness":
            self.framework.check_readiness(args.version)

        if str(args.operation) == "list-phases":
            self.framework.list_phases()

        if str(args.operation) == "list-phase-tasks":
            self.framework.list_phase_tasks(args.version, args.phase)

        if str(args.operation) == "execute-phase":
            self.framework.execute_phase(
                args.version, args.phase, args.skip_task_download)

        if str(args.operation) == "execute":
            self.framework.execute(args.version)

        if str(args.operation) == "extract-zip":
            self.framework.extract_zip(args.zip_path)

        if str(args.operation) == "create-task":
            self.framework.create_task(
                args.phase, args.task_name, args.cls_name)

        if str(args.operation) == "create-shell-script":
            self.framework.create_shell_script(
                args.phase, args.shell_script_name)


class_object = Main()
class_object.main()
